// BBSDlg.cpp : t@C
//

#include "stdafx.h"
#include "PCRPlayer.h"
#include "BBSDlg.h"
#include "afxdialogex.h"

enum TIMER_ID {
	TIMER_ID_POPUP,
	TIMER_ID_CLEAR_TITLE,
	TIMER_ID_CLEAR_STATUS,
	TIMER_ID_CLEAR_POST,
};

// CBBSDlg _CAO

IMPLEMENT_DYNAMIC(CBBSDlg, CDialog)

CBBSDlg::CBBSDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CBBSDlg::IDD, pParent)
	, click_(false)
	, show_(false)
	, icon_(NULL)
	, start_(0)
	, end_(0)
	, posting_(false)
{
	utl::createFont(font_);
	utl::createFont(bold_, true);
	utl::createFont(edit_);
	
	gl_.peca.setBBSCallback([this](){
		if (IsWindow(this->GetSafeHwnd()))
		{
			this->PostMessage(WM_COMMAND, IDM_UPDATE_CHANNEL_INFO_END);
		}
	});

	gl_.bbs.setPostCallback([this](bool result){
		if (IsWindow(this->GetSafeHwnd()))
		{
			if (result)
			{
				this->PostMessage(WM_COMMAND, IDM_UPDATE_POST_SUCCEEDED);
			}
			else
			{
				this->PostMessage(WM_COMMAND, IDM_UPDATE_POST_FAILED);
			}
		}
	});

	gl_.bbs.setUpdateCallback([this](long target){
		if (IsWindow(this->GetSafeHwnd()))
		{
			switch (target)
			{
			case bbs::UPDATE_TITLE:		this->PostMessage(WM_COMMAND, IDM_UPDATE_BBS_TITLE); break;
			case bbs::UPDATE_STATUS:	this->PostMessage(WM_COMMAND, IDM_UPDATE_BBS_STATUS); break;
			case bbs::UPDATE_POST:		this->PostMessage(WM_COMMAND, IDM_UPDATE_BBS_POST); break;

			case bbs::UPDATE_ALL:
			default:
				this->PostMessage(WM_COMMAND, IDM_UPDATE_BBS_ALL); break;
			}
		}
	});
}

CBBSDlg::~CBBSDlg()
{
	if (icon_) { DestroyIcon(icon_); }
	if (font_.GetSafeHandle()) { font_.DeleteObject(); }
	if (bold_.GetSafeHandle()) { bold_.DeleteObject(); }
	if (edit_.GetSafeHandle()) { edit_.DeleteObject(); }

	gl_.peca.setBBSCallback([](){});
	gl_.bbs.setPostCallback([](bool){});
	gl_.bbs.setUpdateCallback([](long){});
}

void CBBSDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_EDIT, m_xcEdit);
	DDX_Control(pDX, IDC_BTN_WRITE, m_xcBtnWrite);
}

void CBBSDlg::setShow(bool show, bool edit)
{
	show_ = show;
	ShowWindow(show ? SW_SHOWNA : SW_HIDE);
	EnableWindow(show ? TRUE : FALSE);
	
	if (edit)
	{
		if (show && !posting_)
		{
			enableControl(true);
			SetFocusEdit();
		}
		else
		{
			enableControl(false);
		}

		if (height_.edit > 0) { return; }
		height_.edit = getEditHeight();
	}
	else
	{
		enableControl(false);

		if (height_.edit <= 0) { return; }
		height_.edit = 0;
	}
}

void CBBSDlg::enableControl(bool enable)
{
	if (enable)
	{
		m_xcEdit.EnableWindow(TRUE);
		m_xcBtnWrite.EnableWindow(TRUE);
		m_xcEdit.SetSel(start_, end_);
		return;
	}

	m_xcEdit.GetSel(start_, end_);
	m_xcEdit.EnableWindow(FALSE);
	m_xcBtnWrite.EnableWindow(FALSE);
}

bool CBBSDlg::enableIfControl(bool enable)
{
	if (enable && IsWindowVisible() && !posting_)
	{
		enableControl(true);
		SetFocusEdit();
		return true;
	}
	enableControl(false);
	return false;
}

void CBBSDlg::SetConfig(const sl::BBSConfig& cfg)
{
	bbs_ = cfg;
	font_.DeleteObject();
	font_.CreateFontIndirect(&bbs_.font);

	bold_.DeleteObject();
	LOGFONT copy = bbs_.font;
	copy.lfWeight = FW_BOLD;
	bold_.CreateFontIndirect(&copy);

	edit_.DeleteObject();
	edit_.CreateFontIndirect(&bbs_.edit.font);



	m_xcEdit.SetFont(&edit_);

	height_.bbs = utl::getFontHeight(GetSafeHwnd(), font_) + (bbs_.marginY * 2);
	height_.edit = gl_.value.window.show.edit ? getEditHeight() : 0;

	Invalidate(FALSE);
}

void CBBSDlg::SetFocusEdit()
{
	// GotoDlgCtrlgƃeLXgSÎŌɖ߂
	int start, end;
	m_xcEdit.GetSel(start, end);
	GotoDlgCtrl(&m_xcEdit);
	m_xcEdit.SetSel(start, end);
}

int CBBSDlg::getEditHeight()
{
	int count = m_xcEdit.GetLineCount();
	if (count <= 1)
	{
		return gl_.value.display.bbs.edit.height;
	}

	int height = utl::getFontHeight(GetSafeHwnd(), *m_xcEdit.GetFont());
	return gl_.value.display.bbs.edit.height + (height * (count - 1));
}

void CBBSDlg::OnSizeHelper()
{
	CRect rc;
	GetClientRect(rc);

	if (IsWindow(m_xcEdit.GetSafeHwnd()))
	{
		m_xcEdit.SetWindowPos(NULL,
			0, rc.Height() - height_.edit,
			rc.Width() - BUTTON_WIDTH, height_.edit,
			SWP_NOZORDER|SWP_NOACTIVATE);
		m_xcEdit.RedrawWindow();
	}

	if (IsWindow(m_xcBtnWrite.GetSafeHwnd()))
	{
		m_xcBtnWrite.SetWindowPos(NULL,
			rc.Width() - BUTTON_WIDTH, rc.Height() - height_.edit,
			BUTTON_WIDTH, height_.edit,
			SWP_NOZORDER|SWP_NOACTIVATE);
		m_xcBtnWrite.RedrawWindow();
	}
}

void CBBSDlg::popup()
{
	res_.setPosition();
	res_.Invalidate(FALSE);

	res_.showWindow(true);
}

BEGIN_MESSAGE_MAP(CBBSDlg, CDialog)
	ON_WM_DESTROY()
	ON_WM_ERASEBKGND()
	ON_WM_PAINT()
	ON_WM_SIZE()
	ON_WM_TIMER()
	ON_WM_MOUSEMOVE()
	ON_WM_MOUSELEAVE()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_LBUTTONDBLCLK()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	ON_COMMAND(IDM_UPDATE_CHANNEL_INFO_END, &CBBSDlg::OnUpdateChannelInfoEnd)
	ON_BN_CLICKED(IDC_BTN_WRITE, &CBBSDlg::OnBnClickedBtnWrite)
	ON_EN_CHANGE(IDC_EDIT, &CBBSDlg::OnEnChangeEdit)
	ON_COMMAND(IDM_UPDATE_POST_SUCCEEDED, &CBBSDlg::OnUpdatePostSucceeded)
	ON_COMMAND(IDM_UPDATE_POST_FAILED, &CBBSDlg::OnUpdatePostFailed)
	ON_COMMAND(IDM_UPDATE_BBS_ALL, &CBBSDlg::OnUpdateBbsAll)
	ON_COMMAND(IDM_UPDATE_BBS_TITLE, &CBBSDlg::OnUpdateBbsTitle)
	ON_COMMAND(IDM_UPDATE_BBS_STATUS, &CBBSDlg::OnUpdateBbsStatus)
	ON_COMMAND(IDM_UPDATE_BBS_POST, &CBBSDlg::OnUpdateBbsPost)
	ON_WM_MBUTTONDOWN()
END_MESSAGE_MAP()


// CBBSDlg bZ[W nh[


BOOL CBBSDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// \ɌȂ悤ɂ
	SetWindowPos(NULL, 0, 0, 0, 0, SWP_NOACTIVATE);

	m_xcEdit.SetWindowText(text_);
	m_xcEdit.SetSel(start_, end_);

	icon_ = (HICON)LoadImage(AfxGetApp()->m_hInstance,
	MAKEINTRESOURCE(IDI_WRITE),
	IMAGE_ICON,
	16, 16,
	LR_DEFAULTCOLOR);

	m_xcBtnWrite.SetIcon(icon_);


	// X|bvAbv
	if (res_.GetSafeHwnd())
	{
		res_.DestroyWindow();
	}

	res_.Create(CResDlg::IDD, this);
	res_.setTarget(this);
	res_.SetWindowPos(NULL, 0, 0, 0, 0, SWP_NOZORDER|SWP_NOACTIVATE);
	//hint_.MoveWindow(0, 0, 300, 18);
	res_.setAlign(DT_LEFT);

	std::vector<bbs::ResInfo> dat;
	gl_.bbs.getDat(dat);
	res_.setDat(dat);

	OnSizeHelper();

	KillTimer(TIMER_ID_CLEAR_STATUS);
	SetTimer(TIMER_ID_CLEAR_STATUS, 3000, NULL);
	KillTimer(TIMER_ID_CLEAR_POST);
	SetTimer(TIMER_ID_CLEAR_POST, 3000, NULL);

	m_xcEdit.SetFont(&edit_);
	m_xcEdit.EnableWindow(posting_ ? FALSE : TRUE);
	m_xcBtnWrite.EnableWindow(posting_ ? FALSE : TRUE);

	return TRUE;  // return TRUE unless you set the focus to a control
	// O : OCX vpeB y[W͕K FALSE Ԃ܂B
}


BOOL CBBSDlg::PreTranslateMessage(MSG* pMsg)
{
	switch (pMsg->message)
	{
	// L[bZ[Weɓn
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		const CWnd* wnd = GetFocus();
		if (wnd && wnd->GetSafeHwnd() == m_xcEdit.GetSafeHwnd())
		{
			DWORD key = static_cast<DWORD>(pMsg->wParam);
			// Shift + Enter ŏ
			if (key == VK_RETURN && (GetAsyncKeyState(VK_SHIFT) & 0x8000))
			{
				OnBnClickedBtnWrite();
				return TRUE;
			}
			else if (key == VK_TAB)
			{// TabL[eɓn
				return FALSE;
			}

			// GfBbgɃtH[JXƂ͐eɃL[nȂ
			break;
		}
		return FALSE;
	}
	return CDialog::PreTranslateMessage(pMsg);
}

void CBBSDlg::OnDestroy()
{
	CDialog::OnDestroy();

	m_xcEdit.GetWindowText(text_);
	m_xcEdit.GetSel(start_, end_);
}


BOOL CBBSDlg::OnEraseBkgnd(CDC* pDC)
{
	return TRUE;
	//return CDialog::OnEraseBkgnd(pDC);
}


void CBBSDlg::OnPaint()
{
	CPaintDC dc(this); // device context for painting
	// TODO: ɃbZ[W nh[ R[hǉ܂B
	// `惁bZ[W CDialog::OnPaint() ĂяoȂłB

	CRect org;
	GetClientRect(&org);

	CDC memDC;
	memDC.CreateCompatibleDC(&dc);

	CBitmap bmp;
	bmp.CreateCompatibleBitmap(&dc, org.Width(), org.Height());
	CBitmap* oldBmp = memDC.SelectObject(&bmp);
	CFont* oldFont = memDC.SelectObject(&font_);
	////////////////////////////////////////////////////////////////////////////////////

	CRect rc(org);

	rc.DeflateRect(0, 0, 0, height_.edit);

	utl::drawStatusBar(memDC, &rc, bbs_.color);

	rc.DeflateRect(bbs_.marginX, 0, bbs_.marginX, 0);

	memDC.SetTextColor(bbs_.text[0]);
	memDC.SetBkColor(RGB(0, 0, 0));
	memDC.SetBkMode(TRANSPARENT);

	auto drawText = [this, &memDC](const std::wstring text, CRect& rc, DWORD align, bool bold){
		CFont* old = NULL;
		if (bold) { old = memDC.SelectObject(&bold_); }
		memDC.DrawText(text.c_str(), -1, rc, DT_NOPREFIX|DT_SINGLELINE|DT_VCENTER|DT_END_ELLIPSIS|align);
		if (bold) { memDC.SelectObject(old); }
	};

	auto calcText = [this, &memDC](const std::wstring text, CRect& rc, DWORD align, bool bold){
		CFont* old = NULL;
		if (bold) { old = memDC.SelectObject(&bold_); }
		memDC.DrawText(text.c_str(), -1, rc, DT_CALCRECT|DT_NOPREFIX|DT_SINGLELINE|DT_VCENTER|align);
		if (bold) { memDC.SelectObject(old); }
	};

	// JE^[
	CRect calc;
	calcText(L"0000", calc, DT_RIGHT, false);
	counter_.width = calc.Width();
	
	int count = 0;
	if (gl_.bbs.getCounter(count))
	{
		drawText((boost::wformat(L"%d") % count).str(), rc, DT_RIGHT, false);
	}
	else
	{
		drawText(L"---", rc, DT_RIGHT, false);
	}
	
	std::wstring title;
	gl_.bbs.getTitle(title);

	std::wstring status;
	gl_.bbs.getStatus(status);
	if (!status.empty()) { status = L" " + status; }

	std::wstring post;
	gl_.bbs.getPost(post);
	if (!post.empty()) { post += L" "; }


	calcText(title, calc, DT_LEFT, bbs_.title.bold);
	int left = max(rc.left, (org.Width() - calc.Width()) / 2);
	int right = min(rc.right - counter_.width, (org.Width() + calc.Width()) / 2);
	right = max(rc.left, right);

	CRect rcCenter(rc);
	rcCenter.left = left;
	rcCenter.right = right;

	//-----------------------------------------------
	calcText(status, calc, DT_LEFT, false);
	left = max(rc.left, rc.right - counter_.width - calc.Width());
	right = max(rc.left, rc.right - counter_.width);

	CRect rcRight(rc);
	rcRight.left = left;
	rcRight.right = right;

	//-----------------------------------------------
	calcText(post, calc, DT_LEFT, false);
	left = rc.left;
	right = min(rc.left + calc.Width(), rcRight.left);

	CRect rcLeft(rc);
	rcLeft.left = left;
	rcLeft.right = right;

	//-----------------------------------------------
	rcCenter.left = max(rcCenter.left, rcLeft.right);
	rcCenter.right = rcRight.left;


	if (gl_.bbs.isStop())
	{
		COLORREF oldColor = memDC.GetTextColor();
		memDC.SetTextColor(bbs_.title.text[0]);
		drawText(title, rcCenter, DT_LEFT, bbs_.title.bold);
		memDC.SetTextColor(oldColor);
	}
	else
	{
		drawText(title, rcCenter, DT_LEFT, bbs_.title.bold);
	}

	drawText(status, rcRight, DT_LEFT, false);
	drawText(post, rcLeft, DT_LEFT, false);


	///////////////////////////////////////////////////////////////
	dc.BitBlt(0, 0, org.Width(), org.Height(), &memDC, 0, 0, SRCCOPY);
	memDC.SelectObject(oldFont);
	memDC.SelectObject(oldBmp);
	bmp.DeleteObject();
}


void CBBSDlg::OnSize(UINT nType, int cx, int cy)
{
	CDialog::OnSize(nType, cx, cy);
	OnSizeHelper();
	Invalidate(FALSE);
}


void CBBSDlg::OnTimer(UINT_PTR nIDEvent)
{
	switch (nIDEvent)
	{
	case TIMER_ID_POPUP:
		{
			popup();
			KillTimer(TIMER_ID_POPUP);
			break;
		}
	case TIMER_ID_CLEAR_TITLE:
		{
			gl_.bbs.setTitle(L"");
			Invalidate(FALSE);
			KillTimer(TIMER_ID_CLEAR_TITLE);
			break;
		}
	case TIMER_ID_CLEAR_STATUS:
		{
			gl_.bbs.setStatus(L"");
			Invalidate(FALSE);
			KillTimer(TIMER_ID_CLEAR_STATUS);
			break;
		}
	case TIMER_ID_CLEAR_POST:
		{
			gl_.bbs.setPost(L"");
			Invalidate(FALSE);
			KillTimer(TIMER_ID_CLEAR_POST);
			break;
		}
	}
	CDialog::OnTimer(nIDEvent);
}


void CBBSDlg::OnMouseMove(UINT nFlags, CPoint point)
{
	auto delayPopup = [this](){
		if (gl_.value.detail.board.delay > 0)
		{
			this->SetTimer(TIMER_ID_POPUP, gl_.value.detail.board.delay, NULL);
		}
		else
		{
			this->popup();
		}
	};

	// }EXĎJn(OnMouseLeave)
	TRACKMOUSEEVENT tme;
	tme.cbSize = sizeof(tme);
	tme.dwFlags = TME_LEAVE;
	tme.hwndTrack = GetSafeHwnd();
	tme.dwHoverTime = HOVER_DEFAULT;
	TrackMouseEvent(&tme);
	
	CRect org;
	GetClientRect(org);

	CRect rc(org);
	rc.DeflateRect(bbs_.marginX, 0, bbs_.marginX, height_.edit);

	CRect crc(rc);
	crc.left = max(rc.left, rc.right - counter_.width);
	if (PtInRect(crc, point))
	{
		SetCursor(LoadCursor(NULL, IDC_HAND));
		delayPopup();
		return;
	}
	
	if (PtInRect(rc, point))
	{
		delayPopup();
		return;
	}

	if (res_.IsWindowVisible())
	{
		res_.showWindow(false);
	}
	
	KillTimer(TIMER_ID_POPUP);
	SetCursor(LoadCursor(NULL, IDC_ARROW));
	
	CDialog::OnMouseMove(nFlags, point);
}


void CBBSDlg::OnMouseLeave()
{
	KillTimer(TIMER_ID_POPUP);
	SetCursor(LoadCursor(NULL, IDC_ARROW));
	res_.showWindow(false);

	CDialog::OnMouseLeave();
}


void CBBSDlg::OnLButtonDown(UINT nFlags, CPoint point)
{
	SetFocus();

	CRect org;
	GetClientRect(org);

	CRect rc(org);
	rc.DeflateRect(bbs_.marginX, 0, bbs_.marginX, height_.edit);

	CRect crc(rc);
	crc.left = max(rc.left, rc.right - counter_.width);
	if (PtInRect(crc, point))
	{
		SetCursor(LoadCursor(NULL, IDC_HAND));
		click_ = true;
		SendMessage(WM_COMMAND, IDM_THREAD_RELOAD);
		return;
	}

	if (PtInRect(rc, point))
	{
		KillTimer(TIMER_ID_POPUP);
		res_.showWindow(false);
		SendMessage(WM_COMMAND, IDM_THREAD);
		return;
	}

	CDialog::OnLButtonDown(nFlags, point);
}


void CBBSDlg::OnLButtonUp(UINT nFlags, CPoint point)
{
	if (click_)
	{
		SetCursor(LoadCursor(NULL, IDC_HAND));
		click_ = false;
	}
	//CDialog::OnLButtonUp(nFlags, point);
}


void CBBSDlg::OnLButtonDblClk(UINT nFlags, CPoint point)
{
	OnLButtonDown(nFlags, point);

	//CDialog::OnLButtonDblClk(nFlags, point);
}

void CBBSDlg::OnRButtonDown(UINT nFlags, CPoint point)
{

	CDialog::OnRButtonDown(nFlags, point);
}


void CBBSDlg::OnRButtonUp(UINT nFlags, CPoint point)
{
	CRect org;
	GetClientRect(org);

	CRect rc(org);
	rc.bottom = rc.top + height_.bbs;

	if (PtInRect(rc, point))
	{
		SendMessage(WM_COMMAND, IDM_BBS_MENU);
		return;
	}

	CDialog::OnRButtonUp(nFlags, point);
}


void CBBSDlg::OnUpdateChannelInfoEnd()
{
	if (gl_.value.detail.board.contact)
	{
		std::wstring url;
		gl_.peca.getContactURL(url);

		//if (url.empty())
		//{// R^NgURL̎ɐݒ肷URL
		//	url = L"";
		//	gl_.latest.contact = url;
		//}

		if (url.empty())
		{
			url = gl_.latest.contact;
		}
		else
		{
			if (url != gl_.latest.contact)
			{
				gl_.latest.contact = url;
				gl_.bbs.execute(url, gl_.value.detail.board);
			}
			else
			{
				if (!gl_.bbs.once(url, gl_.value.detail.board))
				{// łɓǂݍ܂Ă
				 //if (gl_.value.detail.board.reload && gl_.bbs.teminated())
				 //{// XbhXVLXV~
				 //	gl_.bbs.reload();// XbhXVJn
				 //}
				}
			}
		}
	}
}


void CBBSDlg::OnBnClickedBtnWrite()
{
	CString text;
	m_xcEdit.GetWindowText(text);

	std::wstring mail = gl_.value.dialog.bbs.sage ? L"sage" : L"";
	gl_.bbs.post(L"", mail, std::wstring(text));

	posting_ = true;

	enableIfControl(false);
}


void CBBSDlg::OnEnChangeEdit()
{
	int fix = getEditHeight() - height_.edit;

	if (fix == 0) { return; }
	
	height_.edit = getEditHeight();

	CWnd* sub = GetParent();
	if (!sub) { return; }

	CMainDlg* main = (CMainDlg*)sub->GetParent();
	if (!main) { return; }
	main->ResizeSubpane(fix);
}


void CBBSDlg::OnUpdatePostSucceeded()
{
	m_xcEdit.SetWindowText(L"");
	OnEnChangeEdit();

	posting_ = false;
	enableIfControl(true);
}


void CBBSDlg::OnUpdatePostFailed()
{
	posting_ = false;
	enableIfControl(true);
}



void CBBSDlg::OnUpdateBbsAll()
{
	Invalidate(FALSE);
}


void CBBSDlg::OnUpdateBbsTitle()
{
	Invalidate(FALSE);
}


void CBBSDlg::OnUpdateBbsStatus()
{
	Invalidate(FALSE);
	KillTimer(TIMER_ID_CLEAR_STATUS);
	SetTimer(TIMER_ID_CLEAR_STATUS, 3000, NULL);
}


void CBBSDlg::OnUpdateBbsPost()
{
	Invalidate(FALSE);
	KillTimer(TIMER_ID_CLEAR_POST);
	SetTimer(TIMER_ID_CLEAR_POST, 3000, NULL);
}


void CBBSDlg::OnMButtonDown(UINT nFlags, CPoint point)
{
	if (gl_.value.operation.mouse.bbs.transfer)
	{// ʏ㑀ֈϏ
	 // eWɕϊ
		ClientToScreen(&point);
		reinterpret_cast<CSubpaneDlg*>(GetParent())->ScreenToClient(&point);
		reinterpret_cast<CSubpaneDlg*>(GetParent())->OnMButtonDown(nFlags, point);
		return;
	}

	CSubpaneDlg* sub = reinterpret_cast<CSubpaneDlg*>(GetParent());
	if (!sub) { return; }
	CMainDlg* main = reinterpret_cast<CMainDlg*>(sub->GetParent());
	if (!main) { return; }

	main->MouseMClickHelper(gl_.value.operation.mouse.bbs.mclick);

	//CDialog::OnMButtonDown(nFlags, point);
}

