
#include "stdafx.h"

namespace network {

PeerCast::PeerCast()
	: valid_(false)
{
}


PeerCast::PeerCast(const std::wstring& url)
	: valid_(false)
{
	using namespace boost::xpressive;

	try {
		std::wstring rx = L"http:\\/\\/([^\\:]+):([^\\/]+)\\/pls\\/([0-9A-Za-z]{32})";

		wsmatch what;
		if (regex_search(url, what, wsregex::compile(rx)))
		{
			url_ = url;
			host_ = what[1];
			port_ = what[2];
			id_ = what[3];
			valid_ = true;
		}
	} catch (...) {}
}

bool PeerCast::connect(std::wstring& stream, int timeout) const
{
	HttpClient client(url_, DEFAULT_USERAGENT, timeout);
	if (!client) { return false; }

	std::vector<char> buf(1024*32, 0);
	int state = client.read(buf.data(), (unsigned int)buf.size());
	if (state == HttpClient::STATE_ERROR || state == HttpClient::STATE_TIMEOUT)
	{
		return false;
	}

	analyse(toUTF16(std::string(buf.begin(), buf.end())), stream);

	client.close();
	return true;
}


bool PeerCast::command(const std::wstring& cmd) const
{
	if (id_ == L"") { return false; }
	std::wstring url = L"http://" + host_ + L":" + port_ + L"/admin?cmd=" + cmd + L"&id=" + id_;

	HttpClient client(url, DEFAULT_USERAGENT, DEFAULT_TIMEOUT);
	if (client)
	{
		client.close();
		return true;
	}
	return false;
}

bool PeerCast::info(ChannelInfo& info) const
{
	std::wstring url = L"http://" + host_ + L":" + port_ + L"/admin?cmd=viewxml";

	HttpClient client(url, DEFAULT_USERAGENT, DEFAULT_TIMEOUT);
	std::vector<char> data;
	if (client.get(data))
	{
		client.close();

		data.push_back(L'\0');

		bool exist = false;

		try {
			using namespace boost::property_tree;
			// xmlǂݍ
			wptree org;
			boost::property_tree::read_xml(std::wstringstream(toUTF16(data.data())), org);

			XmlParser pt(org);
			auto p = pt.get(L"peercast.channels_relayed");
			p.search(L"channel", [&](const XmlParser& x)->bool{

				std::wstring id;
				x.get(L"id", id);

				if(_wcsicmp(id.c_str(), id_.c_str()) == 0)
				{
					info.id = id;

					x.get(L"name", info.name);
					x.get(L"bitrate", info.bitrate);
					x.get(L"type", info.type);
					x.get(L"genre", info.genre);
					x.get(L"desc", info.desc);
					x.get(L"url", info.url);
					x.get(L"uptime", info.uptime);
					x.get(L"comment", info.comment);
					x.get(L"skips", info.skips);
					x.get(L"age", info.age);
					x.get(L"bcflags", info.bcflags);

					auto relay = x.get(L"relay");
					relay.get(L"listeners", info.relay.listeners);
					relay.get(L"relays", info.relay.relays);
					relay.get(L"hosts", info.relay.hosts);
					relay.get(L"status", info.relay.status);
					relay.get(L"firewalled", info.relay.firewalled);

					auto track = x.get(L"track");
					track.get(L"title", info.track.title);
					track.get(L"artist", info.track.artist);
					track.get(L"album", info.track.album);
					track.get(L"genre", info.track.genre);
					track.get(L"contact", info.track.contact);

					info.cnv.bitrate = wcstol(info.bitrate.c_str(), NULL, 10);
					info.cnv.listeners = wcstol(info.relay.listeners.c_str(), NULL, 10);
					info.cnv.relays = wcstol(info.relay.relays.c_str(), NULL, 10);

					// genre_desc
					if (!info.genre.empty() && !info.desc.empty())
					{
						info.cnv.genreDesc = info.genre + L"-" + info.desc;
					}
					else if (!info.genre.empty())
					{
						info.cnv.genreDesc = info.genre;
					}
					else if (!info.desc.empty())
					{
						info.cnv.genreDesc = info.desc;
					}

					// track
					if (!info.track.artist.empty())
					{
						info.cnv.playing = info.track.artist;
					}

					if (!info.track.title.empty())
					{
						if (!info.cnv.playing.empty()) { info.cnv.playing += L"/"; }
						info.cnv.playing += info.track.title;
					}

					if (!info.track.album.empty())
					{
						if (!info.cnv.playing.empty()) { info.cnv.playing += L"/"; }
						info.cnv.playing += info.track.album;
					}
					exist = true;
					return true;
				}
				return false;
			});
		} catch (...) { return false; }
		return exist;
	}
	return false;
}


bool PeerCast::analyse(const std::wstring& playlist, std::wstring& stream) const
{
	using namespace boost::xpressive;

	try {
		//std::wstring rx = L"(mms|https?):\\/\\/[^\\/]+\\/stream\\/[0-9a-zA-Z]{32}(\\.[_0-9a-zA-Z]+)?";
		//std::wstring rx = L"(mmsh?|https?):\\/\\/[^\\/]+\\/stream\\/[0-9a-zA-Z]{32}(\\.[_0-9a-zA-Z]+)?";
		std::wstring rx = L"[0-9a-z]+:\\/\\/[^\\/]+\\/stream\\/[0-9a-zA-Z]{32}(\\.[_0-9a-zA-Z]+)?";

		wsmatch what;
		if (regex_search(playlist, what, wsregex::compile(rx)))
		{
			stream = what[0];
			return !stream.empty();
		}
	} catch (...) {}
	return false;
}

std::wstring PeerCast::toUTF16(const std::string& src) const
{
	const int size = ::MultiByteToWideChar(CP_UTF8, 0, src.c_str(), -1, NULL, 0);

	if (size < 0) { return std::wstring(); }

	std::vector<wchar_t> buff(size);
	
	int result = ::MultiByteToWideChar(CP_UTF8, 0, src.c_str(), -1, buff.data(), size);

	if (result < 0) { return std::wstring(); }

	return buff.data();
}

} // namespace network