#pragma once

#include <stdarg.h>
#include <boost/thread/recursive_mutex.hpp>
#include <boost/optional.hpp>

namespace utl {

struct Log {
	enum LOGTYPE {
		LOGTYPE_GRAPH,
		LOGTYPE_FILTER,
		LOGTYPE_BBS,
		LOGTYPE_PEERCAST,
		LOGTYPE_NONE,
	};

	int type;
	time_t time;
	std::wstring desc;
	std::wstring detail;
};

class SyncLog {
	boost::recursive_mutex mutex_;
	std::deque<Log> log_;

public:
	SyncLog();
	virtual ~SyncLog();

	void clear();
	void add(int type, const wchar_t* desc, ...);
	void add(int type, const wchar_t* desc, va_list args);
	void add(int type, const std::wstring& desc, const wchar_t* detail, ...);
	void add(int type, const std::wstring& desc, const wchar_t* detail, va_list args);
	bool get(std::deque<Log>& log);
	bool get(std::deque<Log>& log, const boost::function<bool(const Log&)>& pred);
};

} // namespace utl
