#include "stdafx.h"

namespace utl {

bool WindowSnap::snap(int a, int b)
{
	return (abs(a - b) < distance_);
}

WindowSnap::WindowSnap()
	: distance_(10)
	, x_(0)
	, y_(0)
{}

void WindowSnap::OnEnterSizeMove(HWND hwnd)
{
	CRect rc, shape;
	GetWindowRect(hwnd, &rc);

	if (utl::getWindowShapeRect(hwnd, shape))
	{
		diff_.left = abs(rc.left - shape.left);
		diff_.top = abs(rc.top - shape.top);
		diff_.right = abs(rc.right - shape.right);
		diff_.bottom = abs(rc.bottom - shape.bottom);
	}

	rc.DeflateRect(diff_);

	CPoint pt;
	GetCursorPos(&pt);

	x_ = pt.x - rc.left;
	y_ = pt.y - rc.top;
}

void WindowSnap::OnMoving(HWND hwnd, LPRECT pRect, bool screen, bool window)
{
	if (!screen && !window) { return; }

	CRect rc(pRect);

	rc.DeflateRect(diff_);
	CRect org(rc);

	//-------------------

	CPoint pt;
	GetCursorPos(&pt);
	rc.OffsetRect(pt.x - (rc.left + x_), pt.y - (rc.top  + y_));

	if (screen && !window)
	{
		MONITORINFO mi;
		if (utl::getDesktopRectFromRect(rc, mi))
		{
			screenSnapX(hwnd, mi, rc);
			screenSnapY(hwnd, mi, rc);
		}
	}
	else if (!screen && window)
	{
		windowSnapX(hwnd, org, rc);
		windowSnapY(hwnd, org, rc);
	}
	else
	{
		MONITORINFO mi;
		if (utl::getDesktopRectFromRect(rc, mi))
		{
			if (!screenSnapX(hwnd, mi, rc)) { windowSnapX(hwnd, org, rc); }
			if (!screenSnapY(hwnd, mi, rc)) { windowSnapY(hwnd, org, rc); }
		}
		else
		{
			windowSnapX(hwnd, org, rc);
			windowSnapY(hwnd, org, rc);
		}
	}

	//-------------------
	rc.InflateRect(diff_);
	*pRect = rc;
}

bool WindowSnap::screenSnapX(HWND hwnd, const MONITORINFO& mi, CRect& rc)
{
	CRect work(mi.rcWork);
	CRect monitor(mi.rcMonitor);
	if (snap(work.left, rc.left))
	{
		rc.OffsetRect(work.left - rc.left, 0); return true;
	}
	else if (snap(work.right, rc.right))
	{
		rc.OffsetRect(work.right - rc.right, 0); return true;
	}
	else if (snap(monitor.left, rc.left))
	{
		rc.OffsetRect(monitor.left - rc.left, 0); return true;
	}
	else if (snap(monitor.right, rc.right))
	{
		rc.OffsetRect(monitor.right - rc.right, 0); return true;
	}
	return false;
}

bool WindowSnap::screenSnapY(HWND hwnd, const MONITORINFO& mi, CRect& rc)
{
	CRect work(mi.rcWork);
	CRect monitor(mi.rcMonitor);

	if (snap(work.top, rc.top))
	{
		rc.OffsetRect(0, work.top - rc.top); return true;
	}
	else if (snap(work.bottom, rc.bottom))
	{
		rc.OffsetRect(0, work.bottom - rc.bottom); return true;
	}
	else if (snap(monitor.top, rc.top))
	{
		rc.OffsetRect(0, monitor.top - rc.top); return true;
	}
	else if (snap(monitor.bottom, rc.bottom))
	{
		rc.OffsetRect(0, monitor.bottom - rc.bottom); return true;
	}
	return false;
}

class Side {
	HWND hwnd_;
public:
	Side(LONG x, LONG y)
		: hwnd_(WindowFromPoint(CPoint(x, y)))
	{}

	bool getRect(CRect& other)
	{
		if (!hwnd_) { return false; }
		HWND hother = GetAncestor(hwnd_, GA_ROOT);
		if (!hother) { return false; }
		if (utl::getWindowShapeRect(hother, other)) { return true; }
		if (GetWindowRect(hother, other)) { return true; }
		return false;
	}
};

bool WindowSnap::windowSnapX(HWND hwnd, const CRect& org, CRect& rc)
{
	Side left(org.left - distance_, org.top + org.Height() / 2);
	Side right(org.right + distance_, org.top + org.Height() / 2);

	CRect other;
	if (left.getRect(other) && snap(other.right, rc.left))
	{
		rc.OffsetRect(other.right - rc.left, 0); return true;
	}
	else if (right.getRect(other) && snap(other.left, rc.right))
	{
		rc.OffsetRect(other.left - rc.right, 0); return true;
	}
	return false;
}

bool WindowSnap::windowSnapY(HWND hwnd, const CRect& org, CRect& rc)
{
	Side top(org.left + org.Width() / 2, org.top - distance_);
	Side bottom(org.left + org.Width() / 2, org.bottom + distance_);

	CRect other;
	if (top.getRect(other) && snap(other.bottom, rc.top))
	{
		rc.OffsetRect(0, other.bottom - rc.top); return true;
	}
	else if (bottom.getRect(other) && snap(other.top, rc.bottom))
	{
		rc.OffsetRect(0, other.top - rc.bottom); return true;
	}
	return false;
}



};

